@@ -40,7 +40,7 @@ GecodeSolverFactory::GecodeSolverFactory() {
   sc.mznlibVersion(1);
   sc.supportsMzn(false);
   sc.description("Internal Gecode presolver plugin");
-  sc.stdFlags({"-a", "-n"});
+  sc.stdFlags({"-a", "-n", "-p"});
   SolverConfigs::registerBuiltinSolver(sc);
 }
 
@@ -155,6 +155,14 @@ bool GecodeSolverFactory::processOption(SolverInstanceBase::Options* opt, int& i
     _opt.verbose = true;
   } else if (string(argv[i]) == "-s" || string(argv[i]) == "--solver-statistics") {
     _opt.statistics = true;
+  } else if (string(argv[i]) == "-p") {
+    if (++i == argv.size()) {
+      return false;
+    }
+    int threads = atoi(argv[i].c_str());
+    if (threads >= 1) {
+      _opt.threads = threads;
+    }
   } else {
     return false;
   }
@@ -531,9 +539,13 @@ void GecodeSolverInstance::processFlatZinc() {
               } else if (var.isbool()) {
                 currentSpace->bvIntroduced[var.index()] = false;
               } else if (var.isfloat()) {
+#ifdef GECODE_HAS_FLOAT_VARS
                 currentSpace->fvIntroduced[var.index()] = false;
+#endif
               } else if (var.isset()) {
+#ifdef GECODE_HAS_SET_VARS
                 currentSpace->svIntroduced[var.index()] = false;
+#endif
               }
             }
           }
@@ -1333,6 +1345,7 @@ void GecodeSolverInstance::prepareEngine() {
 
     engineOptions.c_d = _opt.c_d;
     engineOptions.a_d = _opt.a_d;
+    engineOptions.threads = _opt.threads;
 
     int seed = _opt.seed;
     double decay = _opt.decay;
@@ -1365,7 +1378,14 @@ void GecodeSolverInstance::printStatistics() {
   Gecode::Search::Statistics stat = engine->statistics();
   auto* solns2out = getSolns2Out();
   StatisticsStream ss(solns2out->getOutput(), solns2out->opt.flagEncapsulateJSON);
-  ss.add("variables", currentSpace->iv.size() + currentSpace->bv.size() + currentSpace->sv.size());
+  auto varcount = currentSpace->iv.size() + currentSpace->bv.size();
+#ifdef GECODE_HAS_SET_VARS
+  varcount += currentSpace->sv.size();
+#endif
+#ifdef GECODE_HAS_FLOAT_VARS
+  varcount += currentSpace->fv.size();
+#endif
+  ss.add("variables", varcount);
   ss.add("propagators", Gecode::PropagatorGroup::all.size(*currentSpace));
   ss.add("propagations", stat.propagate);
   ss.add("nodes", stat.node);
@@ -1842,9 +1862,7 @@ void GecodeSolverInstance::setSearchStrategyFromAnnotation(
       // branchInfo.add(bh,r0,r1,names);
 #else
       if (!ignoreUnknown) {
-        err << "Warning, ignored search annotation: ";
-        flatAnn[i]->print(err);
-        err << std::endl;
+        err << "Warning, ignored search annotation: " << *i << std::endl;
       }
 #endif
     } else if (Expression::isa<Call>(i) &&
@@ -1856,9 +1874,7 @@ void GecodeSolverInstance::setSearchStrategyFromAnnotation(
       def_set_valsel = ann2svalsel(Expression::cast<Id>(call->arg(1))->str(), r0, r0, rnd);
 #else
       if (!ignoreUnknown) {
-        err << "Warning, ignored search annotation: ";
-        flatAnn[i]->print(err);
-        err << std::endl;
+        err << "Warning, ignored search annotation: " << *i << std::endl;
       }
 #endif
     } else if (Expression::isa<Call>(i) &&
